/*
  file:   mod_auth_checkpasswd.c
  author: Sebastian Zagrodzki <s.zagrodzki@net.icm.edu.pl>
  date:   19/05/2001

  This used to be mod_auth.c in the Apache source tree.
*/

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"

#include <stdio.h>

/*
  mod_auth_checkpasswd uses checkpasswd interface (as described in 
  D. J. Bernstein's CHECKPASSWORD document) to authenticate users.
  
  Example .htaccess file
  
  AuthCheckPasswd "/var/qmail/bin/checkpasswd"
  AuthCPGroupFile
  AuthCPAuthoritative

  
*/

extern int errno;

typedef struct auth_config_struct {
    char *auth_pwfile;
    char *auth_grpfile;
    int auth_authoritative;
} auth_config_rec;

static void *create_auth_dir_config(pool *p, char *d)
{
    auth_config_rec *sec =
    (auth_config_rec *) ap_pcalloc(p, sizeof(auth_config_rec));
    sec->auth_pwfile = NULL;
    sec->auth_grpfile = NULL;
    sec->auth_authoritative = 1;
    return sec;
}

static const char *set_auth_slot(cmd_parms *cmd, void *offset, char *f, char *t)
{
    if (t && strcmp(t, "standard"))
	return ap_pstrcat(cmd->pool, "Invalid auth file type: ", t, NULL);

    return ap_set_file_slot(cmd, offset, f);
}

static const command_rec auth_checkpasswd_cmds[] =
{
    {"AuthCheckPasswd", set_auth_slot,
     (void *) XtOffsetOf(auth_config_rec, auth_pwfile), OR_AUTHCFG, TAKE12,
     "application that checks username and password"},
    {"AuthCPGroupFile", set_auth_slot,
     (void *) XtOffsetOf(auth_config_rec, auth_grpfile), OR_AUTHCFG, TAKE12,
     "text file containing group names and member user IDs"},
    {"AuthCPAuthoritative", ap_set_flag_slot,
     (void *) XtOffsetOf(auth_config_rec, auth_authoritative),
     OR_AUTHCFG, FLAG,
     "Set to 'off' to allow access control to be passed along to lower modules if the UserID is not known to this module"},
    {NULL}
};

module MODULE_VAR_EXPORT auth_checkpasswd_module;

static char *get_pw(request_rec *r, char *user, const char* password, char *auth_pwfile) {
    int filedes[2];
    int retval;
    int status;
    time_t timestamp;

    setenv ("REMOTE_ADDR", r -> connection -> remote_ip, 1);
    
    if (pipe(filedes) < 0)
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "Could not create pipe");

    switch (fork()) {
	    case -1:
		    ap_log_rerror (APLOG_MARK, APLOG_ERR, r, "Could not fork");
		    return NULL;
		    break;
	    case 0:
		    dup2(filedes[0], 3);
		    close(filedes[1]);
		    ap_log_rerror(APLOG_MARK, APLOG_WARNING|APLOG_NOERRNO, r, "Trying to exec %s", auth_pwfile);
		    execl(auth_pwfile, auth_pwfile, NULL);
		    ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "Could not exec %s", auth_pwfile);
		    exit(-1);
		    break;
	    default:
		    break;
    }
    
    close(filedes[0]);
    time(&timestamp);
    write(filedes[1], user, strlen(user)+1);
    write(filedes[1], password, strlen(password)+1);
    write(filedes[1], ctime(&timestamp), strlen(ctime(&timestamp))+1);
    close(filedes[1]);
    
    if (wait(&retval) < 0) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "Wait failed");
	    return NULL;
    }

    if (WIFEXITED(retval) == 0) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "Child exited abnormally");
	    return NULL;
    }

    status = WEXITSTATUS(retval);
    ap_log_rerror (APLOG_MARK, APLOG_WARNING|APLOG_NOERRNO, r, "Status: %d", status);

    if (status > 1)
	ap_log_rerror (APLOG_MARK, APLOG_ERR, r, "Checkpassword error: %d", status);
    if (status)
      return NULL;
    else
      return user;

}

static table *groups_for_user(pool *p, char *user, char *grpfile)
{
    configfile_t *f;
    table *grps = ap_make_table(p, 15);
    pool *sp;
    char l[MAX_STRING_LEN];
    const char *group_name, *ll, *w;

    if (!(f = ap_pcfg_openfile(p, grpfile))) {
	aplog_error(APLOG_MARK, APLOG_ERR, NULL,
		    "Could not open group file: %s", grpfile);
	return NULL;
    }

    sp = ap_make_sub_pool(p);

    while (!(ap_cfg_getline(l, MAX_STRING_LEN, f))) {
	if ((l[0] == '#') || (!l[0]))
	    continue;
	ll = l;
	ap_clear_pool(sp);

	group_name = ap_getword(sp, &ll, ':');

	while (ll[0]) {
	    w = ap_getword_conf(sp, &ll);
	    if (!strcmp(w, user)) {
		ap_table_setn(grps, ap_pstrdup(p, group_name), "in");
		break;
	    }
	}
    }
    ap_cfg_closefile(f);
    ap_destroy_pool(sp);
    return grps;
}

static int authenticate_basic_user(request_rec *r)
{
    auth_config_rec *sec =
    (auth_config_rec *) ap_get_module_config(r->per_dir_config, &auth_checkpasswd_module);
    conn_rec *c = r->connection;
    const char *sent_pw;
    char *real_pw;
    char *invalid_pw;
    int res;
    FILE* fp;
    
    if ((res = ap_get_basic_auth_pw(r, &sent_pw)))
	return res;

    if (!sec->auth_pwfile)
	return DECLINED;

    if (!(real_pw = get_pw(r, c->user, sent_pw, sec->auth_pwfile))) {
	if (!(sec->auth_authoritative))
	    return DECLINED;
	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
		      "user %s: Authentication failure from: %s URI: %s",
		      c->user, sec->auth_pwfile, r->uri);
	ap_note_basic_auth_failure(r);
	return AUTH_REQUIRED;
    }

    return OK;
    
}

static int check_user_access(request_rec *r)
{
    auth_config_rec *sec =
    (auth_config_rec *) ap_get_module_config(r->per_dir_config, &auth_checkpasswd_module);
    char *user = r->connection->user;
    int m = r->method_number;
    int method_restricted = 0;
    register int x;
    const char *t, *w;
    table *grpstatus;
    const array_header *reqs_arr = ap_requires(r);
    require_line *reqs;
    FILE* fp;

    if (!reqs_arr)
	return (OK);
    reqs = (require_line *) reqs_arr->elts;

    if (sec->auth_grpfile)
	grpstatus = groups_for_user(r->pool, user, sec->auth_grpfile);
    else
	grpstatus = NULL;

    for (x = 0; x < reqs_arr->nelts; x++) {

	if (!(reqs[x].method_mask & (1 << m)))
	    continue;

	method_restricted = 1;

	t = reqs[x].requirement;
	w = ap_getword_white(r->pool, &t);
	if (!strcmp(w, "valid-user"))
	    return OK;
	if (!strcmp(w, "user")) {
	    while (t[0]) {
		w = ap_getword_conf(r->pool, &t);
		if (!strcmp(user, w))
		    return OK;
	    }
	}
	else if (!strcmp(w, "group")) {
	    if (!grpstatus)
		return DECLINED;

	    while (t[0]) {
		w = ap_getword_conf(r->pool, &t);
		if (ap_table_get(grpstatus, w))
		    return OK;
	    }
	} else if (sec->auth_authoritative) {
	    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
		"access to %s failed, reason: unknown require directive:"
		"\"%s\"", r->uri, reqs[x].requirement);
	}
    }

    if (!method_restricted)
	return OK;

    if (!(sec->auth_authoritative))
	return DECLINED;

    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
	"access to %s failed, reason: user %s not allowed access",
	r->uri, user);
	
    ap_note_basic_auth_failure(r);
    return AUTH_REQUIRED;
}

module MODULE_VAR_EXPORT auth_checkpasswd_module =
{
    STANDARD_MODULE_STUFF,
    NULL,			/* initializer */
    create_auth_dir_config,	/* dir config creater */
    NULL,			/* dir merger --- default is to override */
    NULL,			/* server config */
    NULL,			/* merge server config */
    auth_checkpasswd_cmds,		/* command table */
    NULL,			/* handlers */
    NULL,			/* filename translation */
    authenticate_basic_user,	/* check_user_id */
    check_user_access,		/* check auth */
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
    NULL,			/* logger */
    NULL,			/* header parser */
    NULL,			/* child_init */
    NULL,			/* child_exit */
    NULL			/* post read-request */
};
